#include "llvm/Support/CodeGen.h"
#include "ast.cpp"
#include "llvm/ADT/APInt.h"
#include "llvm/IR/Constants.h"
#include "llvm/IR/IRBuilder.h"
#include "llvm/IR/LLVMContext.h"
#include "llvm/IR/LegacyPassManager.h"
#include "llvm/IR/Module.h"
#include "llvm/Support/FileSystem.h"
#include "llvm/Support/TargetRegistry.h"
#include "llvm/Support/TargetSelect.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetOptions.h"
#include <fstream>

static llvm::LLVMContext llvm_context;
static llvm::IRBuilder<> llvm_builder(llvm_context);
static std::unique_ptr<llvm::Module> llvm_module;

static std::map<std::string, llvm::AllocaInst*> llvm_symbol_table;

llvm::Value* ASTIntegerExpression::codegen()
{
    llvm::APInt llvm_int_value = llvm::APInt(32, value, false);
    return llvm::ConstantInt::get(llvm_context, llvm_int_value);
}

llvm::Value* ASTVariableExpression::codegen()
{
	// TODO check if value exists
	return llvm_builder.CreateLoad(llvm_symbol_table[name], name.c_str());
}

llvm::Value* ASTReturnExpression::codegen()
{
    return llvm_builder.CreateRet(val->codegen());
}

llvm::Value* ASTVariableDeclarationExpression::codegen() {
	// declaration aka allocation
	llvm::Type* type = llvm::Type::getInt32Ty(llvm_context);
	llvm::AllocaInst* alloc_inst = llvm_builder.CreateAlloca(type, 0, name);

	llvm_symbol_table[name] = alloc_inst;

	// initialization aka storing value
	return llvm_builder.CreateStore(val->codegen(), alloc_inst);
}

void write_llvm_file()
{
    std::string Str;
    std::error_code output_stream_error;

    llvm::raw_string_ostream output_stream(Str);

    output_stream << *llvm_module;
    output_stream.flush();

    std::ofstream out("out.ll");
    out << Str;
    out.close();
}

void write_obj_file()
{
    llvm::InitializeAllTargetInfos();
    llvm::InitializeAllTargets();
    llvm::InitializeAllTargetMCs();
    llvm::InitializeAllAsmParsers();
    llvm::InitializeAllAsmPrinters();

    std::string target_triple = llvm::sys::getDefaultTargetTriple();
    llvm_module->setTargetTriple(target_triple);

    std::string Error;
    const llvm::Target* target = llvm::TargetRegistry::lookupTarget(target_triple, Error);

    // Print an error and exit if we couldn't find the requested target.
    // This generally occurs if we've forgotten to initialise the
    // TargetRegistry or we have a bogus target triple.
    if (!target) {
        llvm::errs() << Error;
        return;
    }

    std::string target_cpu_arch = "generic";
    std::string target_features = "";
    llvm::TargetOptions target_options;

    llvm::Optional<llvm::Reloc::Model> reloc_model = llvm::Optional<llvm::Reloc::Model>();

    llvm::TargetMachine* target_machine = target->createTargetMachine(target_triple, target_cpu_arch, target_features, target_options, reloc_model);

    llvm_module->setDataLayout(target_machine->createDataLayout());

    std::string output_filename = "out.o";
    std::error_code output_stream_error;
    llvm::raw_fd_ostream output_stream(output_filename, output_stream_error, llvm::sys::fs::OF_None);

    if (output_stream_error) {
        llvm::errs() << "Could not open file: " << output_stream_error.message();
        return;
    }

    llvm::legacy::PassManager pass;
    llvm::TargetMachine::CodeGenFileType output_file_type = llvm::TargetMachine::CGFT_ObjectFile;

    if (target_machine->addPassesToEmitFile(pass, output_stream, nullptr, output_file_type)) {
        llvm::errs() << "target_machine can't emit a file of this type";
        return;
    }

    pass.run(*llvm_module);
    output_stream.flush();

    llvm::outs() << "Wrote " << output_filename << "\n";
}
