#include "reader.cpp"
#include <iostream>

#define LEX_SUCCESS 0
#define LEX_UNKNOWN 1
#define LEX_END 2

#define TOK_IDENT 0
#define TOK_BRACKET_ROUND_OPEN 1
#define TOK_BRACKET_ROUND_CLOSE 2
#define TOK_RET 3
#define TOK_NUMBER 4
#define TOK_VAR 5

typedef struct token {
    int tok_type;
    std::string tok_data;
    int tok_line; // the line this token is in
    int tok_col; // position of first char in token
} token_t;

int get_token(token_t* tok)
{
    char next_char = peek_char();

    while (isspace(next_char)) {
        pop_char();
        next_char = peek_char();
    }

    if (next_char == EOF) {
        return LEX_END;
    }

    // check for staring words with a letter at the beginning
    if (isalpha(next_char)) {
        std::string identifier(1, pop_char());

        tok->tok_col = input_file_cursor_col;
        tok->tok_line = input_file_cursor_line;

        while (isalnum(peek_char())) {
            identifier += pop_char();
        }

        if (identifier == "ret") {
            tok->tok_type = TOK_RET;
        } else if (identifier == "var") {
            tok->tok_type = TOK_VAR;
        } else {
            tok->tok_type = TOK_IDENT;
        }

        tok->tok_data = identifier;

        return LEX_SUCCESS;
    } else if (isdigit(next_char)) {
        std::string number(1, pop_char());

        tok->tok_col = input_file_cursor_col;
        tok->tok_line = input_file_cursor_line;

        while (isdigit(peek_char())) {
            number += pop_char();
        }

        tok->tok_data = number;
        tok->tok_type = TOK_NUMBER;

        return LEX_SUCCESS;
    } else if (next_char == '(') {
        tok->tok_type = TOK_BRACKET_ROUND_OPEN;
        tok->tok_data = std::string(1, pop_char());
        tok->tok_col = input_file_cursor_col;
        tok->tok_line = input_file_cursor_line;

        return LEX_SUCCESS;
    } else if (next_char == ')') {
        tok->tok_type = TOK_BRACKET_ROUND_CLOSE;
        tok->tok_data = std::string(1, pop_char());
        tok->tok_col = input_file_cursor_col;
        tok->tok_line = input_file_cursor_line;

        return LEX_SUCCESS;
    }

    char unknown_char = pop_char();
    printf("ERROR: Found unexpected char '%c' (%d)\n", unknown_char, unknown_char);

    return LEX_UNKNOWN;
}
